import datetime
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
#timestep_size = datetime.timedelta(seconds=1)
#start_time = datetime.datetime.now().replace(microsecond=0)
from stonesoup.models.clutter import ClutterModel
import numpy as np
from stonesoup.models.measurement.nonlinear import Cartesian2DToBearing
from stonesoup.models.measurement.nonlinear import CartesianToBearingRange
# Stone Soup imports:
from stonesoup.types.state import State, GaussianState
from stonesoup.types.array import StateVector
from stonesoup.types.array import CovarianceMatrix
from stonesoup.models.transition.linear import (
    CombinedLinearGaussianTransitionModel, ConstantVelocity)
#from stonesoup.predictor.particle import ParticlePredictor
#from stonesoup.resampler.particle import SystematicResampler
#from stonesoup.updater.particle import ParticleUpdater
from stonesoup.measures import Mahalanobis
from stonesoup.hypothesiser.distance import DistanceHypothesiser
from stonesoup.dataassociator.neighbour import GNNWith2DAssignment
#from stonesoup.tracker.simple import SingleTargetTracker


start_time = datetime.now().replace(microsecond=0)

from stonesoup.platform.base import MovingPlatform

# Define the initial platform position, in this case the origin
initial_loc = StateVector([[0], [0], [0], [0]])
initial_state = State(initial_loc, start_time)

# Define transition model and position for 3D platform
transition_model = CombinedLinearGaussianTransitionModel(
    [ConstantVelocity(0.0), ConstantVelocity(0.0)])

# create our fixed platform
from stonesoup.platform.base import FixedPlatform

# Import a range rate bearing elevation capable radar
from stonesoup.sensor.radar.radar import RadarBearingRange

# Create a radar sensor
radar_noise_covar = CovarianceMatrix(np.diag(
    np.array([ np.deg2rad(3),  # Bearing
              100.,  # Range
               ])))

# Mount the radar onto the platform
params = ((-100, 100),  # clutter min x and max x
          (-100, 100),  # clutter min y and max y
          )

clutter_model = ClutterModel(
     clutter_rate=0.1,
     distribution=np.random.default_rng().uniform,
     dist_params=params
)
radar = RadarBearingRange   (          ndim_state=4,
                                       position_mapping=(0, 2),
                                       noise_covar=None,
                                       mounting_offset=None,
                                       clutter_model=clutter_model
                                       )



sensor1_transition_model = CombinedLinearGaussianTransitionModel(
    [ConstantVelocity(0.), ConstantVelocity(0)])

sensor1_platform = FixedPlatform(
    State([0, 0, 0, 0]),
    position_mapping=[0, 2],
    sensors=[radar]
)



#from stonesoup.sensor.passive import PassiveElevationBearing

imager_noise_covar = CovarianceMatrix(np.diag(np.array([np.deg2rad(1)])))  # Bearing

# imager mounting offset
imager_mounting_offsets = StateVector([0, 0])  # e.g. wing mounted imaging pod
imager_rotation_offsets = StateVector([0, 0])


# Mount the imager onto the platform
imager = Cartesian2DToBearing   (ndim_state=4,
                                 mapping=(0, 2),
                               noise_covar=np.diag([0, 0]),
                                 translation_offset=None,
                                rotation_offset=None
                                )

from stonesoup.platform.base import FixedPlatform
sensor2_platform = FixedPlatform(
    State([0, 0, 0, 0]),
    position_mapping=[0, 2],
    sensors=[imager]
)


#量测模型
from stonesoup.models.measurement.nonlinear import CartesianToBearingRange
sensor_x = 50  # Placing the sensor off-centre
sensor_y = 0


measurement_model = CartesianToBearingRange(
    ndim_state=4,
    mapping=(0, 2),
    noise_covar=np.diag([np.radians(0.2), 1]),
    translation_offset=np.array([[sensor_x], [sensor_y]])
)






#目标场景设置
from stonesoup.types.groundtruth import GroundTruthPath, GroundTruthState
from ordered_set import OrderedSet
np.random.seed(1991)
truths = OrderedSet()
transition_model = CombinedLinearGaussianTransitionModel([ConstantVelocity(0.005),
                                                          ConstantVelocity(0.005)])
timesteps = [start_time]
#匀速直线设置
target1 = GroundTruthPath([GroundTruthState([0, 1, 0, 1], timestamp=timesteps[0])])
for k in range(1, 21):
    timesteps.append(start_time+timedelta(seconds=k))
    target1.append(GroundTruthState(
        transition_model.function(target1[k-1], noise=True, time_interval=timedelta(seconds=1)),
        timestamp=timesteps[k]))
truths.add(target1)
#设置目标2，10秒匀速直线，10秒匀速转弯
#设置匀速直线
target2 = GroundTruthPath([GroundTruthState([0, 1, 20, -1], timestamp=timesteps[0])])
for k in range(1, 11):
    target2.append(GroundTruthState(
        transition_model.function(target2[k-1], noise=True, time_interval=timedelta(seconds=1)),
        timestamp=timesteps[k]))
_ = truths.add(target2)
#设置匀速转弯
from stonesoup.models.transition.linear import KnownTurnRate
turn_noise_diff_coeffs = np.array([0., 0.])
turn_model = KnownTurnRate(turn_noise_diff_coeffs=turn_noise_diff_coeffs, turn_rate=np.radians(30))
start_time = datetime.now()
for k in range(1, 11):
    initiater=target2[-1]
    target2.append(GroundTruthState(
       turn_model.function(initiater, noise=True, time_interval=timedelta(seconds=1)),
        timestamp=start_time+timedelta(seconds=10+k)))
_ = truths.add(target2)

#静态图
plt.figure(figsize=(10, 8))
plt.plot([state.state_vector[0, 0] for state in target1], [state.state_vector[1, 0] for state in target1], label="目标1")
plt.plot([state.state_vector[0, 0] for state in target2], [state.state_vector[1, 0] for state in target2], label="目标2")
plt.xlabel("X位置")
plt.ylabel("Y位置")
plt.title("目标真实轨迹")
plt.legend()
plt.grid(True)
#plt.show()
#动态图
from stonesoup.plotter import AnimatedPlotterly
plotter = AnimatedPlotterly(timesteps, tail_length=0.3)
plotter.plot_ground_truths(truths, [0, 2])
plt=plotter.fig
#plt.show()

#探测设置
from stonesoup.simulator.platform import PlatformDetectionSimulator

radar_simulator = PlatformDetectionSimulator(
    groundtruth=truths,
    platforms=[sensor1_platform]
)

imager_simulator = PlatformDetectionSimulator(
    groundtruth=truths,
    platforms=[sensor2_platform]
)
from itertools import tee
from stonesoup.feeder.multi import MultiDataFeeder

radar_plotting, radar_simulator, s1_detector = tee(radar_simulator, 3)
s2_detector = tee(imager_simulator, 3)
imager_plotting, imager_simulator, s2_detector = tee(imager_simulator, 3)

s1s2_detector = MultiDataFeeder([radar_simulator, radar_simulator])

#更新
s1_detections = set()
s2_detections = set()


#for (time, s1ds), (_, s2ds) in zip(radar_plotting, imager_plotting):
    ##s2_detections.update(s2ds)

#滤波跟踪设置


#运行跟踪器

from stonesoup.predictor.kalman import KalmanPredictor
predictor=KalmanPredictor(transition_model)
from stonesoup.updater.kalman import ExtendedKalmanUpdater
updater=ExtendedKalmanUpdater(measurement_model=None)

from stonesoup.hypothesiser.distance import DistanceHypothesiser
from stonesoup.measures import Mahalanobis
hypothesiser = DistanceHypothesiser(predictor, updater, measure=Mahalanobis(), missed_distance=3)
#构建数据关联器
from stonesoup.dataassociator.neighbour import GNNWith2DAssignment
data_associator = GNNWith2DAssignment(hypothesiser)
#删除器设置
from stonesoup.deleter.time import UpdateTimeStepsDeleter
deleter = UpdateTimeStepsDeleter(time_steps_since_update=100)

min_detections = 3  # number of detections required to begin a track

initiator_prior_state = GaussianState(
    state_vector=np.array([[0], [0], [0], [0]]),
    covar=np.diag([0, 10, 0, 10])**2
)

noise_covar = CovarianceMatrix(np.array(np.diag([
                                                 np.deg2rad(0.15)**2,
                                                 25**2])))
from stonesoup.models.measurement.nonlinear import CartesianToBearingRange
initiator_meas_model = CartesianToBearingRange(
    ndim_state=4,
    mapping=np.array([0, 2]),
    noise_covar=noise_covar
)

from stonesoup.initiator.simple import MultiMeasurementInitiator
initiator= MultiMeasurementInitiator(
    prior_state=initiator_prior_state,
    measurement_model=measurement_model,
    deleter=deleter,
    data_associator=data_associator,
    updater=updater,
    min_points=min_detections
)
from stonesoup.tracker.pointprocess import PointProcessMultiTargetTracker
from stonesoup.tracker.simple import MultiTargetTracker
#追踪器设置
from copy import deepcopy
from stonesoup.mixturereducer.gaussianmixture import GaussianMixtureReducer
from stonesoup.types.state import TaggedWeightedGaussianState
reducer = GaussianMixtureReducer(
    prune_threshold=1E-3,
    pruning=True,
    merge_threshold=200,
    merging=True
)
birth_covar = CovarianceMatrix(np.diag([10, 1, 10, 1]))
birth_component = TaggedWeightedGaussianState(
    state_vector=[0, 0, 0, 0],
    covar=birth_covar**2,
    weight=0.5,
    tag=TaggedWeightedGaussianState.BIRTH,
    timestamp=start_time
)
meas_fusion_tracker = PointProcessMultiTargetTracker(
    detector=s1s2_detector,
    hypothesiser=deepcopy(hypothesiser),
    updater=deepcopy(updater),
    reducer=deepcopy(reducer),
    birth_component=deepcopy(birth_component),
    extraction_threshold=0.90,
)


tracker = MultiTargetTracker(
    initiator=initiator,
    deleter=deleter,
    detector=s1s2_detector,
    data_associator=data_associator,
    updater=updater
 )


groundtruth = set()
detections = set()
ctracks=set()


for time, ctracks in tracker :
    groundtruth.update(groundtruth_sim.groundtruth_paths)
    detections.update(sim.detections)
    tracks.update(ctracks)







from stonesoup.plotter import AnimatedPlotterly
timesteps = [start_time + timestep_size * k for k in range(number_of_steps)]

plotter = AnimatedPlotterly(timesteps, tail_length=1)
plotter.plot_ground_truths(groundtruth, mapping=[0, 2])
plotter.plot_measurements(detections, mapping=[0, 2])
plotter.plot_tracks(tracks, mapping=[0, 2])
plt=plotter.fig
plt.show()